PROMPT_EXTRA_INDEX = "Would you like to add a corresponding extra-index-url for pip?"
VER_NAMES = "major", "minor", "build"
ERR_NO_NEW_FILES = "ERROR: No New Build, did you forget to update the version?"\
                   " or clear out old trial builds?"
